#ifndef HASHMAP_H
#define HASHMAP_H

#include <stdlib.h>
#include <string.h>

char *strdup(const char *s) {
  size_t len = strlen(s) + 1;
  char *copy = (char*)malloc(len);
  if (copy) {
    memcpy(copy, s, len);
  }
  return copy;
}

#define DEFINE_HASHMAP(StructType, HASHSIZE) \
unsigned int hash_##StructType(char *s) { \
  unsigned int hashval; \
  for (hashval = 0; *s != '\0'; s++) \
    hashval = *s + 31 * hashval; \
  return hashval % HASHSIZE; \
} \
\
StructType *get_##StructType(StructType **map, char *s) { \
  StructType *np; \
  for (np = map[hash_##StructType(s)]; np != NULL; np = np->next) \
    if (strcmp(s, np->keyword) == 0) \
      return np; \
  return NULL; \
} \
\
StructType *put_##StructType(StructType **map, char *keyword, TokenType token) { \
  StructType *np; \
  unsigned int hashval; \
  if ((np = get_##StructType(map, keyword)) == NULL) { \
    np = (StructType *)malloc(sizeof(*np)); \
    if (np == NULL || (np->keyword = strdup(keyword)) == NULL) \
      return NULL; \
    hashval = hash_##StructType(keyword); \
    np->next = map[hashval]; \
    map[hashval] = np; \
  } \
  np->token = token; \
  return np; \
} \
\
int delete_##StructType(StructType **map, char *keyword) { \
  unsigned int hashval = hash_##StructType(keyword); \
  StructType *current = map[hashval]; \
  StructType *prev = NULL; \
  while (current) { \
    if (strcmp(current->keyword, keyword) == 0) { \
      if (prev) { \
        prev->next = current->next; \
      } else { \
        map[hashval] = current->next; \
      } \
      free(current->keyword); \
      free(current); \
      return 1; \
    } \
    prev = current; \
    current = current->next; \
  } \
  return 0; \
} \
\
StructType **new_##StructType() { \
  StructType **map = (StructType **)calloc(HASHSIZE, sizeof(StructType *)); \
  if (!map) return NULL; \
  return map; \
} \
\
void free_##StructType(StructType **map) { \
  if (!map) return; \
  int i; \
  for (i = 0; i < HASHSIZE; i++) { \
    StructType *current = map[i]; \
    while (current) { \
      StructType *next = current->next; \
      free(current->keyword); \
      free(current); \
      current = next; \
    } \
  } \
  free(map); \
}

#endif
